//====- XCoreMachineFuctionInfo.h - XCore machine function info -*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file declares XCore-specific per-machine-function information.
//
//===----------------------------------------------------------------------===//

#ifndef XCOREMACHINEFUNCTIONINFO_H
#define XCOREMACHINEFUNCTIONINFO_H

#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineFrameInfo.h"
#include <vector>
#include <cassert>

namespace llvm {

// Forward declarations
class Function;

class XCoreCall {
private:
  unsigned Type;
protected:
  XCoreCall(unsigned T)
  : Type(T) { }
public:
  enum {
    ExternalSymbolType,
    FunctionType
  };
  unsigned getType() const { return Type; }
  static inline bool classof(const XCoreCall *) { return true; }
};

class XCoreFunctionCall : public XCoreCall {
private:
  const Function *Fn;
public:
  XCoreFunctionCall(const Function *F)
  : XCoreCall(XCoreCall::FunctionType),
    Fn(F) {}
    
  const Function *getFunction() { return Fn; }
  
  static inline bool classof(const XCoreFunctionCall *) { return true; }
  static inline bool classof(const XCoreCall *C) {
    return C->getType() == XCoreCall::FunctionType;
  }
};

class XCoreExternalSymbolCall : public XCoreCall {
private:
  const char *Sym;
public:
  XCoreExternalSymbolCall(const char *S)
  : XCoreCall(XCoreCall::ExternalSymbolType),
    Sym(S){}
    
  const char *getSymbol() { return Sym; }
  
  static inline bool classof(const XCoreExternalSymbolCall *) { return true; }
  static inline bool classof(const XCoreCall *C) {
    return C->getType() == XCoreCall::ExternalSymbolType;
  }
};

/// XCoreFunctionInfo - This class is derived from MachineFunction private
/// XCore target-specific information for each MachineFunction.
class XCoreFunctionInfo : public MachineFunctionInfo {
private:
  bool LRSpillSlotSet;
  int LRSpillSlot;
  int FPSpillSlot;
  int VarArgsFrameIndex;
  std::vector<std::pair<unsigned, CalleeSavedInfo> > SpillLabels;
  std::vector<XCoreCall*> Calls;
  bool HasIndirectCalls;

public:
  XCoreFunctionInfo() :
    LRSpillSlotSet(false),
    LRSpillSlot(0),
    FPSpillSlot(0),
    VarArgsFrameIndex(0),
    HasIndirectCalls(false) {}
  
  XCoreFunctionInfo(MachineFunction &MF) :
    LRSpillSlotSet(false),
    LRSpillSlot(0),
    FPSpillSlot(0),
    VarArgsFrameIndex(0),
    HasIndirectCalls(false) {}
  
  ~XCoreFunctionInfo() {
    for (std::vector<XCoreCall*>::iterator it = Calls.begin(),end = Calls.end();
      it != end; ++it)
    {
      delete(*it);
    }
  }
  
  void setVarArgsFrameIndex(int off) { VarArgsFrameIndex = off; }
  int getVarArgsFrameIndex() const { return VarArgsFrameIndex; }
  
  void setHasIndirectCalls(bool val) { HasIndirectCalls = val; }
  bool getHasIndirectCalls() const { return HasIndirectCalls; }
  
  bool hasLRSpillSlot() { return LRSpillSlotSet; }
  
  void setLRSpillSlot(int off) {
    assert(!LRSpillSlotSet && "LR Spill slot set twice");
    LRSpillSlotSet = true;
    LRSpillSlot = off;
  }
  
  int getLRSpillSlot() const {
    assert(LRSpillSlotSet && "LR Spill slot no set");
    return LRSpillSlot;
  }
  
  void setFPSpillSlot(int off) { FPSpillSlot = off; }
  int getFPSpillSlot() const { return FPSpillSlot; }
  
  std::vector<std::pair<unsigned, CalleeSavedInfo> >&getSpillLabels() {
    return SpillLabels;
  }
  
  const std::vector<XCoreCall*>&getCalls() const {
    return Calls;
  }
  
  void addCall(const char *name) {
    Calls.push_back(new XCoreExternalSymbolCall(name));
  }
  
  void addCall(const Function *f) {
    Calls.push_back(new XCoreFunctionCall(f));
  }
};
} // End llvm namespace

#endif // XCOREMACHINEFUNCTIONINFO_H
