//===-- XCoreMetadata.cpp ---------------------------------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// Collect XCore specific metadata for a module.
//
//===----------------------------------------------------------------------===//

#include "XCoreMetadata.h"

#include "llvm/Constants.h"
#include "llvm/GlobalValue.h"
#include "llvm/GlobalVariable.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"

using namespace llvm;

// Handle the Pass registration stuff necessary to use TargetData's.
namespace {
  RegisterPass<XCoreMetadata> X("xcoremetadata", "XCore Metadata");
}
char XCoreMetadata::ID = 0;

XCoreMetadata::XCoreMetadata()
: ImmutablePass((intptr_t)&ID)
{
}

XCoreMetadata::~XCoreMetadata()
{
}

/// GetTypestring - Returns a type string associated with a global
/// value, or 0 if there is no assoicated type string.
const std::string *XCoreMetadata::GetTypeString(const GlobalValue &GV) {
  std::map<const GlobalValue *, std::string>::const_iterator iter =
      TypeStrings.find(&GV);
  if (iter == TypeStrings.end()) {
    return 0;
  }
  return &iter->second;
}

/// GetTypestring - Returns whether a stack attribute is set on the
/// function. If the stack attribute is set then value if updated.
bool XCoreMetadata::
GetStackAttribute(const Function &F, uint64_t &Value)
{
  std::map<const Function *, uint64_t>::const_iterator iter =
      StackAttributes.find(&F);
  if (iter == StackAttributes.end()) {
    return 0;
  }
  Value = iter->second;
  return true;
}

static GlobalValue *getGlobalValue(Value *V)
{
  GlobalValue *Obj = 0;
  if (ConstantExpr *CE = dyn_cast<ConstantExpr>(V)) {
    if (CE->getOpcode() == Instruction::BitCast) {
      Obj = dyn_cast<GlobalValue>(CE->getOperand(0));
    } else if (CE->getOpcode() == Instruction::GetElementPtr &&
        CE->getNumOperands() == 3 &&
          cast<Constant>(CE->getOperand(1))->isNullValue() &&
          cast<Constant>(CE->getOperand(2))->isNullValue()) {
      Obj = dyn_cast<GlobalValue>(CE->getOperand(0));
    } else {
      return 0;
    }
  } else {
    Obj = dyn_cast<GlobalValue>(V);
  }
  return Obj;
}

void XCoreMetadata::
AnalyzeTypeStrings(Module &M)
{
  GlobalVariable *GV = M.getGlobalVariable("xcore.typestrings");
  if (!GV || !GV->hasInitializer()) return;

  // Should be an array of 'i8*,i8*'
  ConstantArray *InitList = dyn_cast<ConstantArray>(GV->getInitializer());
  if (InitList == 0) return;

  for (unsigned i = 0, e = InitList->getNumOperands(); i != e; ++i) {
    ConstantStruct *CS = dyn_cast<ConstantStruct>(InitList->getOperand(i));
    if (!CS || CS->getNumOperands() < 2)
      continue;
    // Get function or variable
    GlobalValue *Obj = getGlobalValue(CS->getOperand(0));
    if (!Obj)
      continue;
    
    // Get string
    GlobalValue *StrGV = getGlobalValue(CS->getOperand(1));
    if (! StrGV)
      continue;
    
    GlobalVariable *TypeString = dyn_cast<GlobalVariable>(StrGV);
    if (!TypeString || !TypeString->hasInitializer())
      continue;
    
    ConstantArray *Str = dyn_cast<ConstantArray>(TypeString->getInitializer());
    
    // Add typestring
    if (!Str || !Str->isCString())
      continue;
    
    std::string S = Str->getAsString();
    // Remove null byte \0
    S.erase(S.size() - 1);
    TypeStrings.insert(std::make_pair(Obj, S));
  }
}

void XCoreMetadata::
AnalyzeStackAttributes(Module &M)
{
  GlobalVariable *GV = M.getGlobalVariable("xcore.stack");
  if (!GV || !GV->hasInitializer()) return;

  // Should be an array of 'i8*,i64'
  ConstantArray *InitList = dyn_cast<ConstantArray>(GV->getInitializer());
  if (InitList == 0) return;

  for (unsigned i = 0, e = InitList->getNumOperands(); i != e; ++i) {
    ConstantStruct *CS = dyn_cast<ConstantStruct>(InitList->getOperand(i));
    if (!CS || CS->getNumOperands() < 2)
      continue;
    // Get function
    GlobalValue *GV = getGlobalValue(CS->getOperand(0));
    if (!GV)
      continue;
    Function *F = dyn_cast<Function>(GV);
    if (!F)
      continue;
    
    // Get value
    ConstantInt *CI = dyn_cast<ConstantInt>(CS->getOperand(1));
    if (! CI)
      continue;
    
    // Add
    StackAttributes.insert(std::make_pair(F, CI->getZExtValue()));
  }
}

void XCoreMetadata::
AnalyzeModule(Module &M)
{
  AnalyzeTypeStrings(M);
  AnalyzeStackAttributes(M);
}
