//===-- XCoreResourceUseWriter.cpp ------------------------------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// Emit resource usage information.
//
//===----------------------------------------------------------------------===//

#include "XCoreResourceUseWriter.h"
#include "XCoreMachineFunctionInfo.h"
#include "XCoreMetadata.h"
#include "llvm/CodeGen/AsmPrinter.h"
#include "llvm/Target/TargetAsmInfo.h"
#include "llvm/Support/Mangler.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Function.h"

#include <ostream>

using namespace llvm;

XCoreResourceUseWriter::XCoreResourceUseWriter(raw_ostream &OS, AsmPrinter *A,
                         const TargetAsmInfo *T)
: O(OS), Asm(A), TAI(T), Meta(0)
{
}

void 
XCoreResourceUseWriter::EmitResourceSymbol(const char *Name, unsigned Type)
{
  O << Name;
  switch (Type) {
  case StackWords:
    O << ".nstackwords";
    break;
  case MaxThreads:
    O << ".maxthreads";
    break;
  case MaxTimers:
    O << ".maxtimers";
    break;
  case MaxChanends:
    O << ".maxchanends";
    break;
  default:
    assert(0 && "Unexpected type");
    O << ".unhandled";
    break;
  }
}

std::string
XCoreResourceUseWriter::GetResourseSymbolName(const Function *F, unsigned Type)
{
  std::string fname = Asm->Mang->getValueName(F);
  switch (Type) {
  case StackWords:
    return fname + ".stackwords";
  case MaxThreads:
    return fname + ".maxthreads";
  case MaxTimers:
    return fname + ".maxtimers";
  case MaxChanends:
    return fname + ".maxchanends";
  default:
    assert(0 && "Unexpected type");
    return fname + ".unhandled";
  }
}

void 
XCoreResourceUseWriter::EmitResourceSymbol(const std::string &Name,
                                           unsigned Type)
{
  EmitResourceSymbol(Name.c_str(), Type);
}

void 
XCoreResourceUseWriter::EmitResourceSymbol(const Function *F, unsigned Type)
{
  EmitResourceSymbol(Asm->Mang->getValueName(F), Type);
}

bool
XCoreResourceUseWriter::EmitUsage(MachineFunction &MF, unsigned Type)
{
  XCoreFunctionInfo *XFI = MF.getInfo<XCoreFunctionInfo>();
  unsigned FunctionResourceUsage;
  bool IncludeCalls = true;
  if (Type == StackWords) {
    uint64_t Value;
    if (Meta && Meta->GetStackAttribute(*MF.getFunction(), Value)) {
      FunctionResourceUsage = Value;
      IncludeCalls = false;
    } else {
      if (XFI->getHasIndirectCalls()) {    
        return false;
      }
      MachineFrameInfo *MFI = MF.getFrameInfo();
      if (MFI->hasVarSizedObjects())
        return false;
      FunctionResourceUsage = MFI->getStackSize() / 4;
    }
  } else {
    if (XFI->getHasIndirectCalls()) {    
      return false;
    }
    if (Type == MaxThreads) {
      FunctionResourceUsage = 1;
    } else {
      // We don't yet allow the use of hardware resources.
      FunctionResourceUsage = 0;
    }
  }
  //assert(TAI->getSetDirective() && "No set directive");
  //O << TAI->getSetDirective();
  
  switch (MF.getFunction()->getLinkage()) {
  default: assert(0 && "Unknown linkage type!");
  case Function::InternalLinkage:  // Symbols default to internal.
  case Function::PrivateLinkage:
    break;
  case Function::ExternalLinkage:
    O << TAI->getGlobalDirective();
    EmitResourceSymbol(CurrentFnName, Type);
    O << "\n";
    break;
  case Function::LinkOnceLinkage:
  case Function::WeakLinkage:
  case Function::CommonLinkage:
    // TODO Use COMDAT groups for LinkOnceLinkage
    O << TAI->getGlobalDirective();
    EmitResourceSymbol(CurrentFnName, Type);
    O << "\n";
    O << TAI->getWeakDefDirective();
    EmitResourceSymbol(CurrentFnName, Type);
    O << "\n";
    break;
  }
  
  O << "\t.linkset\t";
  EmitResourceSymbol(CurrentFnName, Type);
  O << ",";
  O << FunctionResourceUsage;
  if (IncludeCalls) {
    const std::vector<XCoreCall*> &Calls = XFI->getCalls();
    bool NeedMax = false;
    if (! Calls.empty()) {
      O << " + (";
      if (Type == MaxThreads) {
        O << "(";
      }
      bool NonWeakCallee = false;
      for (std::vector<XCoreCall*>::const_iterator it = Calls.begin(),
        end = Calls.end(); it != end; ++it) {
        if (NeedMax) {
          O << " $M ";
        }
        if (XCoreFunctionCall *C = dyn_cast<XCoreFunctionCall>(*it)) {
          const Function *Callee = C->getFunction();
          EmitResourceSymbol(Callee, Type);
          if (Callee->hasExternalWeakLinkage()) {
            WeakSymbols.insert(GetResourseSymbolName(Callee, Type));
          } else {
            NonWeakCallee = true;
          }
        } else {
          assert(isa<XCoreExternalSymbolCall>(*it) && "Unexpected call type");
          XCoreExternalSymbolCall *C = cast<XCoreExternalSymbolCall>(*it);
          EmitResourceSymbol(C->getSymbol(), Type);
        }
        NeedMax = true;
      }
      if (Type == MaxThreads) {
        // If a callee is ExternalWeak and is not defined at the final link
        // then its maxthreads symbol will be set to 0. In cases where all the
        // callees are ExternalWeak we max the max of the callee maxthreads with
        // 1 to ensure the expression is always greater than or equal to one.
        if (!NonWeakCallee) {
          O << " $M 1";
        }
        O << ") - 1";
      }
      O << ")";
    }
  }
  O << "\n";
  return true;
}

void
XCoreResourceUseWriter::EmitFunction(MachineFunction &MF)
{
  //FIXME duplicate calls should be removed.
  CurrentFnName = Asm->Mang->getValueName(MF.getFunction());
  
  EmitUsage(MF, StackWords);
  EmitUsage(MF, MaxThreads);
  EmitUsage(MF, MaxTimers);
  EmitUsage(MF, MaxChanends);
  
  //FIXME doesn't belong here, move into seperate class
  //Need to use linkset here see bug 3153.
  O << "\t.linkset\t" << CurrentFnName << ".locnochandec, 1\n";
  O << "\t.linkset\t" << CurrentFnName << ".locnoside, 1\n";
}

void
XCoreResourceUseWriter::EndModule()
{
  for (std::set<std::string>::const_iterator it = WeakSymbols.begin(),
    end = WeakSymbols.end(); it != end; ++it) {
    O << TAI->getWeakDefDirective() << *it << "\n";
  }
}
